(function() {


  angular.module('mnQuery').controller('mnQueryController', queryController);

  queryController.$inject = ['$rootScope', '$modal', '$timeout', 'mnQueryService', 'mnHelper'];

  function queryController ($rootScope, $modal, $timeout, mnQueryService, mnHelper) {

    var qc = this;

    //
    // current UI version number
    //

    qc.version = "1.0.4";

    //
    // alot of state is provided by the mnQueryService
    //

    qc.buckets = mnQueryService.buckets;             // buckets on cluster
    qc.gettingBuckets = mnQueryService.gettingBuckets;// busy retrieving?
    qc.updateBuckets = mnQueryService.updateBuckets; // function to update
    qc.lastResult = mnQueryService.getResult();      // holds the current query and result
    qc.limit = mnQueryService.limit;                 // automatic result limiter
    qc.executingQuery = mnQueryService.executingQuery;

    // some functions for handling query history, going backward and forward

    qc.prev = mnQueryService.prevResult;
    qc.next = mnQueryService.nextResult;

    qc.hasNext = mnQueryService.hasNextResult;
    qc.hasPrev = mnQueryService.hasPrevResult;

    qc.getCurrentIndex = mnQueryService.getCurrentIndex;
    qc.clearHistory= mnQueryService.clearHistory;

    // variable and code for managing the choice of output format in different tabs

    qc.selectTab = selectTab;
    qc.isSelected = mnQueryService.isSelected;

    qc.status_success = mnQueryService.status_success;
    qc.status_fail = mnQueryService.status_fail;

    //
    // options for the two editors, query and result
    //

    qc.aceInputLoaded = aceInputLoaded;
    qc.aceInputChanged = aceInputChanged;
    qc.aceOutputLoaded = aceOutputLoaded;
    qc.updateOutputEditorSize = updateOutputEditorSize;

    //
    // functions for running queries and saving results
    //

    qc.query = query;
    qc.save = save;

    //
    // options for the two Ace editors, the input and the output
    //

    qc.aceInputOptions = {
        mode: 'n1ql',
        showGutter: true,
        onLoad: qc.aceInputLoaded,
        onChange: qc.aceInputChanged,
        $blockScrolling: Infinity
    };

    qc.aceOutputOptions = {
        mode: 'json',
        showGutter: true,
        useWrapMode: true,
        onLoad: qc.aceOutputLoaded,
        onChange: qc.updateOutputEditorSize,
        $blockScrolling: Infinity
    };

    //
    // call the activate method for initialization
    //

    activate();

    //
    // change the tab selection
    //

    function selectTab(tabNum) {
      mnQueryService.selectTab(tabNum);
      // select focus after a delay to try and force update of the editor
      $timeout(swapEditorFocus,10);
    };

    //
    // the text editor doesn't update visually if changed when off screen. Force
    // update by focusing on it
    //

    function swapEditorFocus() {
      if (qc.outputEditor) {
        qc.outputEditor.focus();
        qc.inputEditor.focus();
      }
    }

    //
    // manage the ACE code editors for query input and JSON output
    //

    var endsWithSemi = /;\s*$/i;

    function aceInputChanged(e) {
      // show a placeholder when nothing has been typed
      var curSession = qc.inputEditor.getSession();
      var noText = curSession.getValue().length == 0;
      var emptyMessageNode = qc.inputEditor.renderer.emptyMessageNode;

      //console.log("Notext: " +noText + ", emptyMessageNode: " + emptyMessageNode);
      if (noText && !emptyMessageNode) {
    	  emptyMessageNode = qc.inputEditor.renderer.emptyMessageNode = document.createElement("div");
    	  emptyMessageNode.textContent = "Enter a query here...";
    	  emptyMessageNode.className = "ace_invisible ace_emptyMessage";
    	  emptyMessageNode.style.padding = "0 5px";
    	  qc.inputEditor.renderer.scroller.appendChild(emptyMessageNode);
      }
      else if (!noText && emptyMessageNode) {
    	  qc.inputEditor.renderer.scroller.removeChild(emptyMessageNode);
    	  qc.inputEditor.renderer.emptyMessageNode = null;
      }
    	
      qc.inputEditor.$blockScrolling = Infinity;

      // if they hit enter and the query ends with a semicolon, run the query
      if (e[0].action === 'insert' && // they typed something
          e[0].end.column === 0 &&    // and ended up on a new line
          e[0].start.row+1 == e[0].end.row && // and added one line
          e[0].start.column > 0 && // and the previous line wasn't blank
          curSession.getLine(e[0].start.row)[e[0].start.column-1] === ';' &&
          endsWithSemi.test(qc.lastResult.query))
        qc.query();

      // as the query is edited, allow it more vertical space, but max sure it
      // doesn't have fewer than 5 lines or more than 15
      var lines = curSession.getLength();
      var height = Math.min(Math.max(75,lines*17),250);
      $("#query_editor").height(height);

    };

    function aceInputLoaded(_editor) {
      var langTools = ace.require("ace/ext/language_tools");

      _editor.$blockScrolling = Infinity;
      _editor.setOptions({enableBasicAutocompletion: true});
      _editor.setFontSize('13px');
      _editor.renderer.setPrintMarginColumn(false);
      qc.inputEditor = _editor;

      if (/^((?!chrome).)*safari/i.test(navigator.userAgent))
        _editor.renderer.scrollBarV.width = 20; // fix for missing scrollbars in Safari

      var identifierCompleter = {
          getCompletions: function(editor, session, pos, prefix, callback) {
            callback(null,mnQueryService.autoCompleteArray);
          }
      };
      langTools.addCompleter(identifierCompleter);

      focusOnInput();
    };


    function aceOutputLoaded(_editor) {
      //console.log("AceOutputLoaded");
      _editor.$blockScrolling = Infinity;
      _editor.setReadOnly(true);
      _editor.renderer.setPrintMarginColumn(false);

      if (/^((?!chrome).)*safari/i.test(navigator.userAgent))
        _editor.renderer.scrollBarV.width = 20; // fix for missing scrollbars in Safari

      qc.outputEditor = _editor;
      updateOutputEditorSize();
    };

    //
    // called when the JSON output changes. We need to make sure the editor is the correct size,
    // since it doesn't auto-resize
    //

    function updateOutputEditorSize() {
    	var constant = 0;
    	if (navigator.userAgent.match(/safari/i))
    		constant = 10;
    	var editor_size = ($('#whole_ui').height() - $('#query_box').height()
    			- $('#result_header').height() - $('#north_ui').height())
    			- $('#result_summary').height() - constant;
    	$('#result_editor').height(editor_size);
    	//console.log("Editor size: " + editor_size);
    }

    $(window).resize(function() {
    	updateOutputEditorSize();
	  });

    //
    // make the focus go to the input field, so that backspace doesn't trigger
    // the browser back button
    //

    function focusOnInput()  {
      if (qc.inputEditor && qc.inputEditor.focus)
        qc.inputEditor.focus();
    }

    //
    // functions for running queries and saving results to a file
    //

    function query() {
      // get rid of whitespace
      qc.lastResult.query = qc.lastResult.query.trim();

      var queryStr = qc.lastResult.query;
      var hasLimitExpr = /limit\s+\d+\s*;\s*$/gi;
      var startsWithSelectExpr = /^\s*select/gi;
      var hasElement = /^\s*select\s*(distinct)?\s*(raw|element|value)\s*/gi;

      var hasLimit = hasLimitExpr.test(queryStr);
      var startsWithSelect = startsWithSelectExpr.test(queryStr);

      //console.log("has limit: " + hasLimit);
      //console.log("startsWithSelect: " + startsWithSelect);

      // strip out any ending semicolon
      if (endsWithSemi.test(queryStr))
        queryStr = queryStr.replace(endsWithSemi,"");

      // add a limit to all "select" statements by wrapping
      if (startsWithSelect && !hasLimit && !hasElement.test(queryStr)) {

        // handle garbage in the limit dialog
        if (isNaN(Number(mnQueryService.limit.max)) ||
            mnQueryService.limit.max < 1)
          mnQueryService.limit.max = mnQueryService.defaultLimit;

        // wrap the query in a new query with a limit
        queryStr = "select cbq_query_workbench_limit.* from (" + queryStr + ") cbq_query_workbench_limit limit " + mnQueryService.limit.max + ";";
      }

      //console.log("got query:"+queryStr);

      // run the query and show a spinner

      var promise = mnQueryService.executeQuery(queryStr);

      if (promise) {
        // for long queries, show a spinner
        mnHelper.promiseHelper(this,promise/*, dialog*/).showSpinner("queryInProgress");
        // also have the input grab focus at the end
        promise.success(focusOnInput)
        .error(focusOnInput);
      }
    };

    //
    // save the results to a file. Here we need to use a scope to to send the file name
    // to the file name dialog and get it back again.
    //

    var dialogScope = $rootScope.$new(true);
    dialogScope.file = {name: "data.json"};

    function save() {
      var isSafari = /^((?!chrome).)*safari/i.test(navigator.userAgent);

      // safari does'nt support saveAs
      if (isSafari) {
        var file = new Blob([qc.lastResult.result],{type: "text/json", name: "data.json"});
        saveAs(file,dialogScope.file.name);
        return;
      }

      // but for those that do, get a name for the file
      var promise = $modal.open({
        templateUrl: 'query/file_dialog/mn_query_file_dialog.html',
        scope: dialogScope
      }).result;

      // now save it
      promise.then(function (res) {
        //console.log("Promise, file: " + tempScope.file.name + ", res: " + res);
        var file = new Blob([qc.lastResult.result],{type: "text/json"});
        saveAs(file,dialogScope.file.name);
      });
    };


    //
    // let's start off with a list of the buckets
    //

    function activate() {
    	// Prevent the backspace key from navigating back. Thanks StackOverflow!
    	$(document).unbind('keydown').bind('keydown', function (event) {
    	    var doPrevent = false;
    	    if (event.keyCode === 8) {
    	        var d = event.srcElement || event.target;
    	        if ((d.tagName.toUpperCase() === 'INPUT' &&
    	             (
    	                 d.type.toUpperCase() === 'TEXT' ||
    	                 d.type.toUpperCase() === 'PASSWORD' ||
    	                 d.type.toUpperCase() === 'FILE' ||
    	                 d.type.toUpperCase() === 'SEARCH' ||
    	                 d.type.toUpperCase() === 'EMAIL' ||
    	                 d.type.toUpperCase() === 'NUMBER' ||
    	                 d.type.toUpperCase() === 'DATE' )
    	             ) ||
    	             d.tagName.toUpperCase() === 'TEXTAREA') {
    	            doPrevent = d.readOnly || d.disabled;
    	        }
    	        else {
    	            doPrevent = true;
    	        }
    	    }

    	    if (doPrevent) {
    	        event.preventDefault();
    	    }
    	});
    	
    	// get the list of buckets
    	qc.updateBuckets();
    }
  }


})();